""" Generator functions to evaluate trial and test functions

NB: Basis evaluation is only needed for the trial function argument in jacobians, as the
multiplication with the test function is part of the sum factorization kernel.
"""
import itertools

from dune.perftool.generation import (backend,
                                      domain,
                                      get_backend,
                                      get_counted_variable,
                                      get_counter,
                                      get_global_context_value,
                                      iname,
                                      instruction,
                                      kernel_cached,
                                      temporary_variable,
                                      )
from dune.perftool.sumfact.tabulation import (basis_functions_per_direction,
                                              construct_basis_matrix_sequence,
                                              BasisTabulationMatrix,
                                              quadrature_points_per_direction,
                                              PolynomialLookup,
                                              name_polynomials,
                                              polynomial_degree,
                                              )
from dune.perftool.sumfact.quadrature import quadrature_inames
from dune.perftool.sumfact.switch import (get_facedir,
                                          get_facemod,
                                          )
from dune.perftool.pdelab.argument import name_coefficientcontainer
from dune.perftool.pdelab.geometry import (local_dimension,
                                           world_dimension,
                                           )
from dune.perftool.loopy.buffer import initialize_buffer, get_buffer_temporary
from dune.perftool.sumfact.symbolic import SumfactKernel, SumfactKernelInputBase
from dune.perftool.options import get_option
from dune.perftool.pdelab.driver import FEM_name_mangling
from dune.perftool.pdelab.restriction import restricted_name
from dune.perftool.pdelab.spaces import name_lfs, name_lfs_bound, lfs_child, name_leaf_lfs
from dune.perftool.tools import maybe_wrap_subscript
from dune.perftool.pdelab.basis import shape_as_pymbolic
from dune.perftool.sumfact.accumulation import sumfact_iname

from ufl.functionview import select_subelement
from ufl import VectorElement, TensorElement

from pytools import product, ImmutableRecord

from loopy.match import Writes

import pymbolic.primitives as prim


class LFSSumfactKernelInput(SumfactKernelInputBase, ImmutableRecord):
    def __init__(self,
                 coeff_func=None,
                 coeff_func_index=None,
                 element=None,
                 component=None,
                 restriction=0,
                 ):
        ImmutableRecord.__init__(self,
                                 coeff_func=coeff_func,
                                 coeff_func_index=coeff_func_index,
                                 element=element,
                                 component=component,
                                 restriction=restriction,
                                 )

    def realize(self, sf, index, insn_dep):
        lfs = name_lfs(self.element, self.restriction, self.component)
        sub_element = select_subelement(self.element, self.component)
        shape = sub_element.value_shape() + (self.element.cell().geometric_dimension(),)

        if isinstance(sub_element, (VectorElement, TensorElement)):
            # Could be 0 but shouldn't be None
            assert self.coeff_func_index is not None

            lfs_pym = lfs_child(lfs,
                                (self.coeff_func_index,),
                                shape=shape_as_pymbolic(shape[:-1]),
                                symmetry=self.element.symmetry())

        leaf_element = sub_element
        if isinstance(sub_element, (VectorElement, TensorElement)):
            leaf_element = sub_element.sub_elements()[0]

        lfs = name_leaf_lfs(leaf_element, self.restriction)
        basisiname = sumfact_iname(name_lfs_bound(lfs), "basis")
        container = self.coeff_func(self.restriction)
        if isinstance(sub_element, (VectorElement, TensorElement)):
            from dune.perftool.pdelab.argument import pymbolic_coefficient as pc
            coeff = pc(container, lfs_pym, basisiname)
        else:
            from dune.perftool.pdelab.argument import pymbolic_coefficient as pc
            coeff = pc(container, lfs, basisiname)

        # Get the input temporary!
        name = get_buffer_temporary(sf.buffer,
                                    shape=(product(mat.basis_size for mat in sf.matrix_sequence), sf.vector_width),
                                    name="input_{}".format(sf.buffer)
                                    )

        assignee = prim.Subscript(prim.Variable(name),
                                  (prim.Variable(basisiname),) + (index,))
        instruction(assignee=assignee,
                    expression=coeff,
                    depends_on=sf.insn_dep.union(insn_dep),
                    tags=frozenset({"sumfact_stage{}".format(sf.stage)}),
                    )

    @property
    def direct_input(self):
        if get_option("fastdg"):
            return self.coeff_func(self.restriction)
        else:
            return None


def _basis_functions_per_direction(element, component):
    """Number of basis functions per direction of a given component of an element"""
    assert len(component.indices()) <= 1
    if len(component.indices()) == 0:
        degree = element.degree()
    else:
        index = component.indices()[0]._value
        degree = element.sub_elements()[index].degree()
    basis_size = degree + 1
    return basis_size


@kernel_cached
def pymbolic_coefficient_gradient(element, restriction, component, coeff_func, visitor_indices):
    rawname = "gradu" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)

    # Number of basis functions per direction
    basis_size = _basis_functions_per_direction(element, component)

    # Get a temporary for the gradient
    from ufl.functionview import select_subelement
    sub_element = select_subelement(element, component)
    rank = len(sub_element.value_shape()) + 1
    shape = sub_element.value_shape() + (world_dimension(),)
    shape_impl = ('arr',) * rank
    temporary_variable(name, shape=shape, shape_impl=shape_impl)

    if len(visitor_indices) == 1:
        coeff_func_index = None
        grad_index, = visitor_indices
    else:
        grad_index, coeff_func_index = visitor_indices

    # Construct the matrix sequence for this sum factorization
    matrix_sequence = construct_basis_matrix_sequence(derivative=grad_index,
                                                      facedir=get_facedir(restriction),
                                                      facemod=get_facemod(restriction),
                                                      basis_size=basis_size,
                                                      )

    inp = LFSSumfactKernelInput(coeff_func=coeff_func,
                                coeff_func_index=coeff_func_index,
                                element=element,
                                component=component,
                                restriction=restriction,
                                )

    # The sum factorization kernel object gathering all relevant information
    sf = SumfactKernel(matrix_sequence=matrix_sequence,
                       preferred_position=grad_index,
                       input=inp,
                       )

    from dune.perftool.sumfact.vectorization import attach_vectorization_info
    vsf = attach_vectorization_info(sf)

    from dune.perftool.sumfact.realization import realize_sum_factorization_kernel
    var, insn_dep = realize_sum_factorization_kernel(vsf)

    return prim.Subscript(var, vsf.quadrature_index(sf)), None


@kernel_cached
def pymbolic_coefficient(element, restriction, component, coeff_func, visitor_indices):
    assert visitor_indices is None
    # Basis functions per direction
    basis_size = _basis_functions_per_direction(element, component)

    # Construct the matrix sequence for this sum factorization
    matrix_sequence = construct_basis_matrix_sequence(facedir=get_facedir(restriction),
                                                      facemod=get_facemod(restriction),
                                                      basis_size=basis_size)

    inp = LFSSumfactKernelInput(coeff_func=coeff_func,
                                element=element,
                                component=component,
                                restriction=restriction,
                                )

    sf = SumfactKernel(matrix_sequence=matrix_sequence,
                       input=inp,
                       )

    from dune.perftool.sumfact.vectorization import attach_vectorization_info
    vsf = attach_vectorization_info(sf)

    # Add a sum factorization kernel that implements the evaluation of
    # the basis functions at quadrature points (stage 1)
    from dune.perftool.sumfact.realization import realize_sum_factorization_kernel
    var, _ = realize_sum_factorization_kernel(vsf)

    return prim.Subscript(var, vsf.quadrature_index(sf)), None


@iname
def sumfact_lfs_iname(bound, dim):
    name = "sumfac_lfs_{}".format(dim)
    domain(name, bound)
    return name


@backend(interface="lfs_inames", name="sumfact")
def lfs_inames(element, restriction, number=1, context=''):
    assert number == 1
    dim = world_dimension()
    return tuple(sumfact_lfs_iname(basis_functions_per_direction(), d) for d in range(dim))


@backend(interface="evaluate_basis")
@kernel_cached
def evaluate_basis(element, name, restriction):
    temporary_variable(name, shape=())
    quad_inames = quadrature_inames()
    inames = lfs_inames(element, restriction)
    facedir = get_facedir(restriction)

    # Collect the pairs of lfs/quad inames that are in use
    # On facets, the normal direction of the facet is excluded
    prod = tuple(BasisTabulationMatrix().pymbolic((prim.Variable(i), prim.Variable(j)))
                 for (i, j) in zip(quad_inames, tuple(iname for i, iname in enumerate(inames) if i != facedir))
                 )

    # Add the missing direction on facedirs by evaluating at either 0 or 1
    if facedir is not None:
        # TODO: Does not work for systems!
        degree = polynomial_degree()

        facemod = get_facemod(restriction)
        prod = prod + (prim.Call(PolynomialLookup(name_polynomials(degree), False),
                                 (prim.Variable(inames[facedir]), facemod)),)

    # Issue the product
    instruction(expression=prim.Product(prod),
                assignee=prim.Variable(name),
                forced_iname_deps=frozenset(quad_inames + inames),
                forced_iname_deps_is_final=True,
                )


def pymbolic_basis(element, restriction, number):
    assert number == 1
    assert element.num_sub_elements() == 0

    name = "phi_{}".format(FEM_name_mangling(element))
    name = restricted_name(name, restriction)
    evaluate_basis(element, name, restriction)

    return prim.Variable(name)


@backend(interface="evaluate_grad")
@kernel_cached
def evaluate_reference_gradient(element, name, restriction, index):
    dim = world_dimension()
    temporary_variable(name, shape=())
    quad_inames = quadrature_inames()
    inames = lfs_inames(element, restriction)
    facedir = get_facedir(restriction)

    # Map the direction to a quadrature iname
    quadinamemapping = {}
    i = 0
    for d in range(local_dimension()):
        if d == facedir:
            i = i + 1
        quadinamemapping[i] = quad_inames[d]
        i = i + 1

    prod = []
    for i in range(dim):
        if i != facedir:
            prod.append(BasisTabulationMatrix(derivative=index == i).pymbolic((prim.Variable(quadinamemapping[i]), prim.Variable(inames[i]))))

    if facedir is not None:
        # TODO: Does not work for systems!
        degree = polynomial_degree()

        facemod = get_facemod(restriction)
        prod.append(prim.Call(PolynomialLookup(name_polynomials(degree), index == facedir),
                              (prim.Variable(inames[facedir]), facemod)),)

    instruction(assignee=prim.Variable(name),
                expression=prim.Product(tuple(prod)),
                forced_iname_deps=frozenset(quad_inames + inames),
                forced_iname_deps_is_final=True,
                )


def pymbolic_reference_gradient(element, restriction, number, indices):
    assert number == 1
    assert len(indices) == 1
    index, = indices

    # TODO: Change name?
    name = "js_{}".format(FEM_name_mangling(element))
    name = restricted_name(name, restriction)
    name = "{}_{}".format(name, index)
    evaluate_reference_gradient(element, name, restriction, index)

    return prim.Variable(name), None
